#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <png.h>

uint32_t read_32bitBE(uint8_t *c)
{
    uint32_t v = 0;
    for (int i = 0; i < 4; i++)
    {
        v *= 0x100;
        v += c[i];
    }
    return v;
}

uint16_t read_16bitBE(uint8_t *c)
{
    uint16_t v = c[0];
    v *= 0x100;
    v += c[1];

    return v;
}

uint32_t read_32bitLE(uint8_t *c)
{
    uint32_t v = 0;
    for (int i = 4-1; i >= 0; i--)
    {
        v *= 0x100;
        v += c[i];
    }
    return v;
}

void process_texture(const char *filename, FILE *infile, long palette_offset, long texture_offset, int idx);
void output_png(const char *filename, const uint8_t *tex, const uint8_t *pal, int idx, unsigned int line_width, unsigned int lines);

int main(int argc, char ** argv)
{
    if (argc != 2)
    {
        printf("gtex2png 0.0\n");
        printf("usage: %s input.gfo3\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    FILE *infile = fopen(argv[1], "rb");
    if (!infile)
    {
        fprintf(stderr, "error opening %s\n", argv[1]);
        goto fail;
    }

    // locate GTEX chunk
    {
        uint8_t chunkbuf[4];
        int found = 0;
        while (4==fread(chunkbuf, 1, 4, infile))
        {
            if (!memcmp(chunkbuf,"GTEX",4))
            {
                found = 1;
                break;
            }

            if (4 != fread(chunkbuf, 1, 4, infile)) goto fail;
            if (read_32bitBE(chunkbuf) < 8) goto fail;
            if (0 != fseek(infile, read_32bitBE(chunkbuf)-8, SEEK_CUR)) goto fail;
        }

        if (!found)
        {
            fprintf(stderr, "no GTEX chunk\n");
            goto fail;
        }
    }

    // collect textures
    {
        uint8_t buf[4];
        int pal_count, tex_count;
        if (4 != fread(buf, 1, 4, infile)) goto fail;
        long gtex_size = read_32bitBE(&buf[0]);
        if (4 != fread(buf, 1, 4, infile)) goto fail;
        pal_count = read_16bitBE(&buf[0]);
        tex_count = read_16bitBE(&buf[2]);
        long gtex_offset = ftell(infile);

        if (pal_count != 1 && pal_count != tex_count)
        {
            fprintf(stderr, "palette count (%d) != texture count (%d)\n", pal_count, tex_count);
            goto fail;
        }

        printf("%d palette%s, %d texture%s\n",
            pal_count, (pal_count==1?"":"s"), tex_count, tex_count==1?"":"s");

        for (int i = 0; i < tex_count; i++)
        {
            long pal_offset;
            long tex_offset;

            if (pal_count == 1)
            {
                if (0 != fseek(infile, gtex_offset, SEEK_SET)) goto fail;
            }
            else
            {
                if (0 != fseek(infile, gtex_offset + i * 4, SEEK_SET)) goto fail;
            }
            if (4 != fread(buf, 1, 4, infile)) goto fail;
            pal_offset = read_32bitBE(buf);
            if (0 != fseek(infile, gtex_offset + (pal_count + i) * 4, SEEK_SET)) goto fail;
            if (4 != fread(buf, 1, 4, infile)) goto fail;
            tex_offset = read_32bitBE(buf);

            process_texture(argv[1], infile, pal_offset, tex_offset, i);
        }
    }

    fclose(infile);
    printf("Done!\n");

    return 0;
fail:
    fprintf(stderr, "failed.\n");
    exit(EXIT_FAILURE);
}

// unlz77wii_raw30 from DSDecmp: http://code.google.com/p/dsdecmp/
// ported to C by Luigi Auriemma
int unlz77wii_raw30(const uint8_t *in, long insz, uint8_t *outdata, long decomp_size) {
    long         curr_size = 0;
    const uint8_t *inl = in + insz;

    while (in < inl)
    {
        uint8_t flag;
        int rl;
        uint8_t compressed;

        // get tag
        flag = *in++;
        compressed = (flag & 0x80);
        rl = flag & 0x7F;
        if (compressed)
            rl += 3;
        else
            rl += 1;
        if (compressed)
        {
            uint8_t b;
            if(in >= inl) goto fail;
            if(curr_size+rl > decomp_size) goto fail;
            b = *in++;
            for (int i = 0; i < rl; i++)
            {
                outdata[curr_size++] = b;
            }
        }
        else
        {
            if(in >= inl) goto fail;
            if(curr_size+rl > decomp_size) goto fail;
            for (int i = 0; i < rl; i++) {
                outdata[curr_size++] = *in++;
            }
        }
    }

    if (curr_size != decomp_size || in != inl)
    {
        fprintf(stderr,"%lx != %lx\n",(unsigned long)curr_size,(unsigned long)decomp_size);
        goto fail;
    }

    return(curr_size);

fail:
    fprintf(stderr,"decompression did not go as expected\n");
    exit(EXIT_FAILURE);
}

void process_texture(const char *filename, FILE *infile, long palette_offset, long texture_offset, int idx)
{
    uint8_t buf[0x14];
    uint8_t palette_buf[0x200];

    // read in palette header
    if (0 != fseek(infile, palette_offset, SEEK_SET)) goto fail;
    if (4 != fread(buf, 1, 4, infile)) goto fail;
    if (read_16bitBE(&buf[0]) != 0 || read_16bitBE(&buf[2]) != 0) goto fail;
    // read in palette
    if (0x200 != fread(palette_buf, 1, 0x200, infile)) goto fail;

    // read texture header
    if (0 != fseek(infile, texture_offset, SEEK_SET)) goto fail;
    if (0xc != fread(buf, 1, 0xc, infile)) goto fail;
    int width = read_16bitBE(&buf[0]);
    int height = read_16bitBE(&buf[2]);
    if (1 != read_32bitBE(&buf[4])) goto fail;
    uint32_t texture_size = read_32bitBE(&buf[8]);

    // read GFCP header
    if (0x14 != fread(buf, 1, 0x14, infile)) goto fail;
    if (memcmp(&buf[0], "GFCP", 4)) goto fail;
    if (1 != read_32bitLE(&buf[4])) goto fail;
    if (2 != read_32bitLE(&buf[8])) goto fail; // compression type = RLE
    uint32_t data_size = read_32bitLE(&buf[0xc]);
    uint32_t compressed_data_size = read_32bitLE(&buf[0x10]);
    if (texture_size-0x14 != compressed_data_size) goto fail;

    // read in, decompress texture
    uint8_t *compressed_texture_buf = malloc(compressed_data_size);
    uint8_t *texture_buf = malloc(data_size);
    if (!compressed_texture_buf || !texture_buf) goto fail;
    if (compressed_data_size != fread(compressed_texture_buf, 1, compressed_data_size, infile)) goto fail;
    unlz77wii_raw30(compressed_texture_buf, compressed_data_size, texture_buf, data_size);
    free(compressed_texture_buf);

    // write PNG
    output_png(filename, texture_buf, palette_buf, idx, width, height);

    free(texture_buf);

    return;

fail:
    fprintf(stderr, "failed processing texture %d\n", idx);
    exit(EXIT_FAILURE);
}

void user_warning_fn(png_structp png, png_const_charp msg)
{
    fprintf(stderr, "PNG Warning: %s\n", msg);
}
void user_error_fn(png_structp png, png_const_charp msg)
{
    fprintf(stderr, "PNG Error: %s\n", msg);
}

void output_png(const char *filename, const uint8_t *tex, const uint8_t *pal, int idx, unsigned int line_width, unsigned int lines)
{
    size_t namelen = strlen(filename) + 2 + 1 + 4 + 1 + 10;
    char *outname = malloc(namelen);
    if (!outname) goto fail;
    snprintf(outname, namelen, "%s_%02d.png", filename, idx);

    FILE *outfile = fopen(outname, "wb");
    const unsigned int tile_width = 8;
    const unsigned int tile_height = 4;

    unsigned int real_line_width = line_width;
    line_width = (line_width+tile_width-1)/tile_width*tile_width;

    uint8_t *linebuf = malloc(line_width*
            ((lines+tile_height-1)/tile_height*tile_height)*4);
    uint8_t **line_pointers = malloc(sizeof(uint8_t*)*lines);

    if (!outfile || !tex || !pal || !linebuf || !line_pointers) goto fail;

    printf("%s: %dx%d\n", outname, real_line_width, lines);

    // Set Up PNG
    png_structp png_ptr = png_create_write_struct
        (PNG_LIBPNG_VER_STRING, (png_voidp)NULL, user_error_fn, user_warning_fn);
    if (!png_ptr) goto fail;

    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr) goto fail;

    if (setjmp(png_jmpbuf(png_ptr)))
    {
        png_destroy_write_struct(&png_ptr, &info_ptr);
        goto fail;
    }

    png_init_io(png_ptr, outfile);

    png_set_IHDR(png_ptr, info_ptr, real_line_width, lines,
        8, PNG_COLOR_TYPE_RGBA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

    // set up palette
    uint8_t palette[256][4];
    {
        for (unsigned int i = 0; i < 256; i++)
        {
            const uint8_t *col = &pal[i*2];

            // RGB5A3, figured out thanks to tpu's brres source
            if (col[0] & 0x80)
            {
                palette[i][0] = (col[0]&0x7C)<<1;
                palette[i][1] = ((col[0]&0x3)<<6)|((col[1]&0xe0)>>2);
                palette[i][2] = ((col[1]&0x1f)<<3);
                palette[i][3] = 0xff;
            }
            else
            {
                palette[i][0] = ((col[1]&0xf)<<4);
                palette[i][1] = ((col[1]&0xf0));
                palette[i][2] = ((col[0]&0xf)<<4);
                palette[i][3] = (col[0]&0x70)<<1;
            }
        }
    }

    // detile and lookup
    for (unsigned int line = 0; line < lines; line += tile_height)
    {
        for (unsigned int col = 0; col < line_width; col += tile_width)
        {
            for (unsigned int y = 0; y < tile_height; y++)
            {
                for (unsigned int x = 0; x < tile_width; x++)
                {
                    uint8_t * const dest_pixel = &linebuf[((line+y)*line_width+col+x)*4];
                    uint8_t src_pixel = tex[line*line_width+col*tile_height+y*tile_width+x];

                    dest_pixel[0] = palette[src_pixel][0];
                    dest_pixel[1] = palette[src_pixel][1];
                    dest_pixel[2] = palette[src_pixel][2];
                    dest_pixel[3] = palette[src_pixel][3];
                }
            }
        }
    }

    // Do PNG
    for (unsigned int line = 0; line < lines; line ++)
    {
        line_pointers[line] = &linebuf[line*line_width*4];
    }
    png_set_rows(png_ptr, info_ptr, line_pointers);
    png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);

    png_destroy_write_struct(&png_ptr, &info_ptr);
    if (0 != fclose(outfile)) goto fail;

    free(linebuf);
    free(line_pointers);

    free(outname);

    return;
fail:
    fprintf(stderr, "failed while trying to generate PNG.\n");
    exit(EXIT_FAILURE);
}
